// -*- tab-width: 4; indent-tabs-mode: nil -*-
#ifndef HDNUM_VECTOR_HH
#define HDNUM_VECTOR_HH

#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>

#include "countablearray.hh"
#include "exceptions.hh"

/** @file
 *  @brief This file implements a generic and dynamic vector class
 */

namespace hdnum {

  /** @brief A flexible vector class 

      with the following features:

       - arbitrary field type
       - dynamic allocation
       - copy with reference semantics
       - subvector extraction
       - randam access and iterator interface
       - full set of operators
       - gnuplot output
       - serialization
   */
  template<class T>
  class Vector {

	typedef CountableArray<T> ArrayType;

  public:

	//! Remember the storage type 
	typedef T value_type;

	/** \brief Reference to an object */
	typedef value_type& reference;

	/** \brief Const reference to an object */
	typedef const value_type& const_reference;

	/** \brief Type used for array indices */
	typedef std::size_t size_type;

	/** \brief Difference type */
	typedef std::ptrdiff_t difference_type;

	//! make empty vector
	Vector () : base()
	{
	  n = 0;
	  data = 0;
	}

	//! make vector with _n uninitialized components
	Vector (size_type _n) : base(new ArrayType(_n))
	{
	  n = _n;
	  data = &((*base)[0]);
	}

	//! make vector with _n initialized components
	Vector (size_type _n, const T& _t) : base(new ArrayType(_n,_t))
	{
	  n = _n;
	  data = &((*base)[0]);
	}

    //! copy constructor with reference semantics
	Vector (const Vector& v) : base(v.base)
    {
      n = v.n;
      data = v.data;
    }

    //! assignment operator with reference semantics
	Vector& operator= (const Vector& v)
    {
      base = v.base;
      data = v.data;
      n = v.n;
    }

    //! make a vector that is identical to components i ... i+m-1
    Vector<T> sub (size_type i, size_type m) const
    {
      Vector<T> v;
      v.base = base;
      v.data = data+i;
      v.n = m;
      return v;
    }

	//! reallocate vector to given size, any data is lost
	void resize (size_type _n)
	{
	  if (n==_n) return; // size already ok
      if (base->get_reference_counter()==1 && data==(&((*base)[0])) && n==base->size())
        {
          base->resize(_n);
          n = _n;
          data = &((*base)[0]);
        }
      else
        HDNUM_THROW(InvalidStateException,"resize called on subvector or while subvectors exist"); 
	}

	//! reallocate vector to given size, any data is lost
	void resize (size_type _n, const T& _t)
	{
	  if (n==_n) return; // size already ok
      if (base->get_reference_counter()==1 && data==(&((*base)[0])) && n==base->size())
        {
          base->resize(_n,_t);
          n = _n;
          data = &((*base)[0]);
        }
      else
        HDNUM_THROW(InvalidStateException,"resize called on subvector or while subvectors exist"); 
	}

	//! Component access
	reference operator[] (size_type i)
	{
	  return data[i];
	}

	//! Const component access
	const_reference operator[] (size_type i) const
	{
	  return data[i];
	}

	//! get array size
	size_type size () const
	{
	  return n;
	}

    //! get index field width for pretty-printing
    int iwidth () const
    {
      return iwidth_;
    }

    //! get data field width for pretty-printing
    int width () const
    {
      return width_;
    }

    //! get data precision for pretty-printing
    int precision () const
    {
      return precision_;
    }

    //! set index field width for pretty-printing
    void iwidth (int i) const
    {
      iwidth_=i;
    }

    //! set data field width for pretty-printing
    void width (int i) const
    {
      width_=i;
    }

    //! set data precision for pretty-printing
    void precision (int i) const
    {
      precision_=i;
    }

    //! random access iterator
    class iterator {
	private:        
		T* p;       
	public:         
      iterator()
      {
        p=0;
      }
      iterator (T* p_)
      {
        p=p_;
      }
      bool operator!= (iterator x) const
      {
        return p!=x.p;
      }
      bool operator== (iterator x) const
      {
        return p==x.p;
      }
      iterator operator++ ()
      {
        p++;
        return *this;
      }
      iterator operator++ (int)
      {
        p++;
        return *this;
      }
      iterator operator-- ()
      {
        p--;
        return *this;
      }
      iterator operator-- (int)
      {
        p--;
        return *this;
      }
      T& operator[] (size_type i) const
      {
        return p[i];
      }
      T& operator* () const
      {
        return *p;
      }
      T* operator-> () const
      {
        return p;
      }
	} ;

    //! const random access iterator
    class const_iterator {
	private:        
		T* p;       
	public:         
      const_iterator()
      {
        p=0;
      }
      const_iterator (T* p_)
      {
        p=p_;
      }
      bool operator!= (const_iterator x) const
      {
        return p!=x.p;
      }
      bool operator== (const_iterator x) const
      {
        return p==x.p;
      }
      const_iterator operator++ ()
      {
        p++;
        return *this;
      }
      const_iterator operator++ (int)
      {
        p++;
        return *this;
      }
      const_iterator operator-- ()
      {
        p--;
        return *this;
      }
      const_iterator operator-- (int)
      {
        p--;
        return *this;
      }
      const T& operator[] (size_type i) const
      {
        return p[i];
      }
      const T& operator* () const
      {
        return *p;
      }
      const T* operator-> () const
      {
        return p;
      }
	} ;

    //! begin iterator
    iterator begin ()
    {
      return iterator(data);
    }
    
	//! end iterator
	iterator end ()
	{
	  return iterator(data+n);
	}
    
    //! begin iterator
    iterator rbegin ()
    {
      return iterator(data+n-1);
    }
    
	//! end iterator
	iterator rend ()
	{
	  return iterator(data-1);
	}


    //! const begin iterator
    const_iterator begin () const
    {
      return const_iterator(data);
    }
    
	//! const end iterator
	const_iterator end () const
	{
	  return const_iterator(data+n);
	}
    
    //! const begin iterator
    const_iterator rbegin () const
    {
      return const_iterator(data+n-1);
    }
    
	//! const end iterator
	const_iterator rend () const
	{
	  return const_iterator(data-1);
	}

    //! vector addition
    Vector& operator+= (const Vector& y)
    {
      for (size_type i=0; i<n; ++i) 
        (*this)[i] += y[i];
      return *this;
    }

    //! vector subtraction
    Vector& operator-= (const Vector& y)
    {
      for (size_type i=0; i<n; ++i) 
        (*this)[i] -= y[i];
      return *this;
    }

    //! vector scalar multiplication
    Vector& operator*= (const T& s)
    {
      for (size_type i=0; i<n; ++i) 
        (*this)[i] *= s;
      return *this;
    }

    //! vector divide by scalar
    Vector& operator/= (const T& s)
    {
      for (size_type i=0; i<n; ++i) 
        (*this)[i] /= s;
      return *this;
    }

    //! add scalar times vector
    Vector& update (const T& a, const Vector& y)
    {
      for (size_type i=0; i<n; ++i) 
        (*this)[i] += a*y[i];
      return *this;
    }

    //! scalar multiplication
    T operator* (const Vector& y)
    {
      T sum(T(0));
      for (size_type i=0; i<n; ++i) 
        sum += (*this)[i]*y[i];
      return sum;
    }

  private:
	CP<ArrayType> base; // array referenced by this vector
	size_type n;        // size of the vector
	T* data;            // pointer to first element used in base
    static int iwidth_, width_, precision_;
  };

  template<class T>
  int Vector<T>::iwidth_ = 4;
  template<class T>
  int Vector<T>::width_ = 15;
  template<class T>
  int Vector<T>::precision_ = 6;

  //! Output operator for Vector
  template <class T>
  inline std::ostream& operator<< (std::ostream& s, const Vector<T>& x)
  {
    for (typename Vector<T>::size_type i=0; i<x.size(); i++)
      s << "[" << std::setw(x.iwidth()) << i << "] " 
        << std::setw(x.width()) << std::scientific << std::showpoint 
        << std::setprecision(x.precision()) << x[i] << std::endl;
    return s;
  }

  //! make a true copy of a vector
  template<class T>
  inline Vector<T> copy (const Vector<T>& v)
  {
    Vector<T> w(v.size());
    for (typename Vector<T>::size_type i=0; i<v.size(); i++)
      w[i] = v[i];
    return w;
  }

  //! gnuplot output for Vector
  template<class T>
  inline void gnuplot (const std::string& fname, const Vector<T>& v)
  {
    std::fstream f(fname.c_str(),std::ios::out); 
    for (typename Vector<T>::size_type i=0; i<v.size(); i++)
      f << std::setw(v.width())  << i << std::scientific << std::showpoint 
        << std::setprecision(v.precision()) << v[i] << std::endl;
    f.close();
  }

  //! gnuplot output for two Vectors
  template<class T>
  inline void gnuplot (const std::string& fname, const Vector<T>& x, const Vector<T>& y)
  {
    std::fstream f(fname.c_str(),std::ios::out); 
    for (typename Vector<T>::size_type i=0; i<x.size(); i++)
      f << std::scientific << std::showpoint 
        << std::setprecision(x.precision()) << x[i] 
        << " "
        << std::scientific << std::showpoint 
        << std::setprecision(y.precision()) << y[i] 
        << std::endl;
    f.close();
  }

  //! fill vector
  template<class T>
  inline void fill (Vector<T>& x, const T& t)
  {
    for (typename Vector<T>::size_type i=0; i<x.size(); i++)
      x[i] = t;
  }

  //! fill vector
  template<class T>
  inline void fill (Vector<T>& x, const T& t, const T& dt)
  {
    T myt(t);
    for (typename Vector<T>::size_type i=0; i<x.size(); i++)
      {
        x[i] = myt;
        myt += dt;
      }
  }

  //! zero out vector
  template<class T>
  inline void zero (Vector<T>& x)
  {
    for (typename Vector<T>::size_type i=0; i<x.size(); i++)
      x[i] = T(0);
  }

} // namespace hdnum

#endif
